using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;

using System.IO;
using System.Globalization;

namespace Barbar.MapReport.Common {
  /// <summary>
  /// Implementace ctvercovho ohranicen geografickho polygonu
  /// </summary>
  [XmlTypeAttribute(Namespace = Constants.XmlNamespace)]
  [Serializable]
  public class GeoBound : IGeoBound {
    private IGeoPoint m_NorthEast;
    private IGeoPoint m_SouthWest;


    public static GeoBound ComputeSharedGeoBound<T>(IEnumerable<T> bounds) where T : IGeoBound {
      GeoBound result = null;
      foreach (T bound in bounds) {
        if (result == null)
          result = new GeoBound(bound);
        if (result.Northeast.Latitude < bound.Northeast.Latitude)
          result.Northeast.Latitude = bound.Northeast.Latitude;
        if (result.Southwest.Latitude > bound.Southwest.Latitude)
          result.Southwest.Latitude = result.Southwest.Latitude;

        if (result.Northeast.Longitude < bound.Northeast.Longitude)
          result.Northeast.Longitude = bound.Northeast.Longitude;
        if (result.Southwest.Longitude > bound.Southwest.Longitude)
          result.Southwest.Longitude = bound.Southwest.Longitude;
      }

      return result;
    }

    /// <summary>
    /// Konstruktor
    /// </summary>
    public GeoBound() {
    }

    /// <summary>
    /// Copy konstruktor
    /// </summary>
    /// <param name="bound">Geografick oblka</param>
    public GeoBound(IGeoBound bound)
      : this(new GeoPoint(bound.Northeast), new GeoPoint(bound.Southwest)) {
    }

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="northeast">Bod nejsevernej a nejvchodnej</param>
    /// <param name="southwest">Bod nejjinej a nejzpadnej</param>
    public GeoBound(IGeoPoint northeast, IGeoPoint southwest) {
      m_NorthEast = new GeoPoint(northeast);
      m_SouthWest = new GeoPoint(southwest);
    }

    public override string ToString() {
      return String.Format(CultureInfo.InvariantCulture, "{0} - {1}", m_SouthWest, m_NorthEast);
    }

    public static bool ContainsPoint(IGeoBound bound, IGeoPoint point) {
      return point.Latitude <= bound.Northeast.Latitude &&
        point.Latitude >= bound.Southwest.Latitude &&
        point.Longitude <= bound.Northeast.Longitude &&
        point.Longitude >= bound.Southwest.Longitude;
    }

    /// <summary>
    /// Vrt true pokud geo oblka obsahuje dan bod
    /// </summary>
    /// <param name="point">GPS bod</param>
    /// <returns></returns>
    public bool ContainsPoint(IGeoPoint point) {
      return GeoBound.ContainsPoint(this, point);
    }


    /// <summary>
    /// Zaserializuje obsah trdy do retezce
    /// </summary>
    /// <returns></returns>
    public string SerializeToString() {
      XmlSerializer serializer = new XmlSerializer(typeof(GeoBound), Constants.XmlNamespace);
      StringBuilder builder = new StringBuilder();
      using (StringWriter writer = new StringWriter(builder, CultureInfo.InvariantCulture)) {
        serializer.Serialize(writer, this);
        writer.Close();
      }
      return builder.ToString();
    }

    /// <summary>
    /// Deserializuje obsah trdy z retezce
    /// </summary>
    /// <param name="value"></param>
    /// <returns></returns>
    public static GeoBound DeserializeFromString(string value) {
      XmlSerializer serializer = new XmlSerializer(typeof(GeoBound), Constants.XmlNamespace);
      GeoBound result = null;
      using (StringReader reader = new StringReader(value)) {
        result = (GeoBound)serializer.Deserialize(reader);
        reader.Close();
      }
      return result;
    }

    /// <summary>
    /// Bod uprostred oblky
    /// </summary>
    [XmlIgnore]
    public GeoPoint Center {
      get { return new GeoPoint((m_NorthEast.Latitude + m_SouthWest.Latitude) / 2, (m_NorthEast.Longitude + m_SouthWest.Longitude) / 2); }
    }

    #region IGeoBound Members

    /// <summary>
    /// nejservernej/nejzpadnej souradnice
    /// </summary>
    public IGeoPoint Northeast {
      get { return m_NorthEast; }
      set { Northeast = value; }
    }

    /// <summary>
    /// nejjinej/nejvchodnej souradnice
    /// </summary>
    public IGeoPoint Southwest {
      get { return m_SouthWest; }
      set { m_SouthWest = value; }
    }

    #endregion

    public GeoBound Resize(decimal percentage) {
      decimal latitude = Math.Abs(m_NorthEast.Latitude - m_SouthWest.Latitude);
      decimal longitude = Math.Abs(m_SouthWest.Longitude - m_NorthEast.Longitude);

      latitude *= percentage;
      longitude *= percentage;

      GeoPoint center = Center;
      GeoPoint northeast = new GeoPoint(center.Latitude + latitude / 2.0m, center.Longitude + longitude / 2.0m);
      GeoPoint southwest = new GeoPoint(center.Latitude - latitude / 2.0m, center.Longitude - longitude / 2.0m);

      return new GeoBound(northeast, southwest);
    }

    public static bool Intersect(IGeoBound firstBound, IGeoBound secondBound) {
      BoundD firstBoundD = new BoundD(firstBound);
      BoundD secondBoundD = new BoundD(secondBound);
      return BoundD.Intersect(firstBoundD, secondBoundD);
    }
  }
}
