using System;
using System.Collections.Generic;
using System.Text;

namespace Barbar.MapReport.Common {
  /// <summary>
  /// Trda reprezentujc Mercantorovu projekci (cylindrick projekce) pouvanou v Google Maps API
  /// </summary>
  public class MercatorProjection {
    private const int TILE_SIZE = 256;
    private double m_Radius;
    private double m_Circumference;
    private double m_FalseEasting;
    private double m_FalseNorthing;
    public static MercatorProjection DefaultProjection = new MercatorProjection(4); // max

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="zoom">rove zveten mapy. Google Maps API podporuj nekolik rovn zveten. V rovni 0 je svet tvoren ctvercem 256x256 pixelu. V rovni 1 je svet tvoren ctyrmi takovmi ctverci, v rovni 2 esncti atd.</param>
    public MercatorProjection(int zoom) {
      m_Circumference = Math.Pow(2.0, zoom) * TILE_SIZE;
      m_Radius = m_Circumference / (2.0 * Math.PI);
      m_FalseEasting = -1.0 * m_Circumference / 2.0;
      m_FalseNorthing = m_Circumference / 2.0;
    }

    /// <summary>
    /// Pretransformuje bod zadan GPS do rovinnho prostoru
    /// </summary>
    /// <param name="point">GPS souradnice bodu</param>
    /// <returns>Bod v rovine</returns>
    public PointD ToPointD(IGeoPoint point) {
      return new PointD(LongitudeToX((double)point.Longitude), LatitudeToY((double)point.Latitude));
    }

    /// <summary>
    /// Pretransformuje bod v rovinnm prostoru do GPS
    /// </summary>
    /// <param name="point">Bod v rovine</param>
    /// <returns>GPS souradnice bodu</returns>
    public GeoPoint ToGeoPoint(PointD point) {
      return new GeoPoint((decimal)YToLatitude(point.Y), (decimal)XToLongitude(point.X));
    }

    /// <summary>
    /// Transformuje zemepisnou dlku (vchod-zpad) do rovinho prostoru
    /// </summary>
    /// <param name="longitudeDegrees">Zemepisn dlka (vchod-zpad)</param>
    /// <returns>Vrt zemepisnou dlku transformovanou do rovinho prostoru</returns>
    public double LongitudeToX(double longitudeDegrees) {
      double longitude = longitudeDegrees * Math.PI / 180.0;
      return m_Radius * longitude - m_FalseEasting;
    }

    /// <summary>
    /// Transformuje x-souradnici rovinnho prostoru na zemepisnou dlku (vchod-zpad)
    /// </summary>
    /// <param name="x">Souradnice</param>
    /// <returns>Vrt zemepisnou dlku</returns>
    public double XToLongitude(double x) {
      double longitude = (x + m_FalseEasting) / m_Radius;
      return longitude * 180.0 / Math.PI;
    }

    /// <summary>
    /// Transformuje zemepisnou rku (sever-jih) do rovinho prostoru
    /// </summary>
    /// <param name="latitudeDegrees">Zemepisn rka (sever-jih)</param>
    /// <returns>Vrt zemepisnou rku transformovanou do rovinho prostoru</returns>
    public double LatitudeToY(double latitudeDegrees) {
      double latitude = latitudeDegrees * Math.PI / 180.0;
      double y = m_Radius / 2.0 *
              Math.Log((1.0 + Math.Sin(latitude)) /
                        (1.0 - Math.Sin(latitude)));
      return m_FalseNorthing - y;
    }

    /// <summary>
    /// Transformuje y-souradnici rovinnho prostoru na zemepisnou rku (sever-jih)
    /// </summary>
    /// <param name="y">Souradnice</param>
    /// <returns>Vrt zemepisnou rku</returns>
    public double YToLatitude(double y) {
      double result;
      y = m_FalseNorthing - y;
      result = (Math.PI / 2 -  ((2 * Math.Atan(Math.Exp(-1.0 * y / m_Radius))))) * 180.0 / Math.PI;
      return result;
    }
  }
}
