using System;
using System.Collections.Generic;
using System.Text;
using Barbar.MapReport.Configuration;
using Microsoft.Practices.EnterpriseLibrary.Caching;
using System.Globalization;
using Barbar.MapReport.DataSource.Resources;
using System.Collections;
using System.Collections.ObjectModel;
using System.Configuration;
using System.Configuration.Provider;
using Barbar.MapReport.DataSource.Query;

namespace Barbar.MapReport.DataSource {
  /// <summary>
  /// Kolekce podkladovch dat
  /// </summary>
  public static class DataSourceManager {
    private static object SyncRoot = new object();
    private static RegionProviderCollection _RegionProviders = new RegionProviderCollection();
    private static RegionGeometryProviderCollection _RegionGeometryProviders = new RegionGeometryProviderCollection();
    private static ZoomLevelProvider _ZoomLevelProvider;
    private static bool _IsInitialized;

    private static void EnsureInitialization() {
      lock (SyncRoot) {
        if (!_IsInitialized)
          Initialize();
      }
    }

    private static ProviderBase InstantiateProvider(ProviderSettings settings) {
      if (settings == null)
        throw new ArgumentNullException("settings");

      Type type = Type.GetType(settings.Type, true);
      
      ProviderBase provider = Activator.CreateInstance(type) as ProviderBase;
      if (provider == null)
        throw new DataSourceException(Translator.Error_ProviderNotCreated);
      provider.Initialize(settings.Name, settings.Parameters);

      return provider;
    }

    private static IEnumerable<ProviderBase> InstantiateProviders(ProviderSettingsCollection settings) {
      if (settings == null)
        throw new ArgumentNullException("settings");

      foreach (ProviderSettings currentSettings in settings) {
        yield return InstantiateProvider(currentSettings);
      }
    }

    private static void Initialize() {
      MapSettingsSection mapSettings = MapConfigurationManager.MapSettings;

      foreach (ProviderBase provider in InstantiateProviders(mapSettings.RegionProvider.Providers)) {
        _RegionProviders.Add(provider);
      }

      foreach (ProviderBase provider in InstantiateProviders(mapSettings.RegionGeometryProvider.Providers)) {
        _RegionGeometryProviders.Add(provider);
      }

      ProviderSettings settings = mapSettings.ZoomLevelProvider.Providers[mapSettings.ZoomLevelProvider.DefaultProvider];
      _ZoomLevelProvider = InstantiateProvider(settings) as ZoomLevelProvider;

      _IsInitialized = true;
    }

    public static Region GetRegionById(string regionId) {
      EnsureInitialization();
      return _RegionProviders.GetRegionById(regionId);
    }

    public static Polygon GetRegionGeometry(string regionId, ZoomLevel zoomLevel) {
      EnsureInitialization();
      return _RegionGeometryProviders.GetRegionGeometry(regionId, zoomLevel);
    }

    public static IEnumerable<Region> GetRegions(RegionCriteria criteria) {
      if (criteria == null)
        throw new ArgumentNullException("criteria");

      EnsureInitialization();
      return _RegionProviders.GetRegions(criteria);
    }

    public static IEnumerable<Region> Regions {
      get {
        EnsureInitialization();
        return _RegionProviders.Regions;
      }
    }

    public static IList<Region> RootRegions {
      get {
        EnsureInitialization();
        return _RegionProviders.RootRegions;
      }
    }

    public static ZoomLevelCollection ZoomLevels {
      get {
        EnsureInitialization();
        return _ZoomLevelProvider.ZoomLevels;
      }
    }

  }
}
